import { BN } from 'ethereumjs-util'
import Web3 from 'web3'
import { FT_ABI, FT_LOCKER_ABI } from './Contracts'
import { Contract } from 'web3-eth-contract'
import { Account } from 'web3-core'

export class ERC20Reactor {
  private web3: Web3
  private contract: Contract
  private locker: Contract
  private account: Account
  private lockerAddress: string

  constructor(web3: Web3, address: string, lockerAddress?: string) {
    this.web3 = web3
    this.account = this.web3.eth.accounts.wallet[0]
    this.contract = new this.web3.eth.Contract(FT_ABI, address, { from: this.account.address })
    if (lockerAddress) {
      this.lockerAddress = lockerAddress
      this.locker = new this.web3.eth.Contract(FT_LOCKER_ABI, lockerAddress, { from: this.account.address })
    }
  }

  /**
   * Get balance or count for current account on specific asset contract.
   *
   * @param address - Asset ERC20 contract address.
   * @param selectedAddress - Current account public address.
   * @returns Promise resolving to BN object containing balance for current account on specific asset contract.
   */
  async getBalanceOf({ address, selectedAddress }: { address?: string; selectedAddress: string }): Promise<BN> {
    const contract = address
      ? new this.web3.eth.Contract(FT_ABI, address, { from: this.account.address })
      : this.contract
    return new Promise<BN>((resolve, reject) => {
      contract.methods.balanceOf(selectedAddress).call({ from: selectedAddress }, (error: Error, result: BN) => {
        /* istanbul ignore if */
        if (error) {
          reject(error)
          return
        }
        resolve(result)
      })
    })
  }

  /**
   * Query for the decimals for a given ERC20 asset.
   *
   * @param address - ERC20 asset contract string.
   * @returns Promise resolving to the 'decimals'.
   */
  async getTokenDecimals(address?: string): Promise<string> {
    const contract = address
      ? new this.web3.eth.Contract(FT_ABI, address, { from: this.account.address })
      : this.contract
    return new Promise<string>((resolve, reject) => {
      contract.methods.decimals().call((error: Error, result: BN | string) => {
        /* istanbul ignore if */
        if (error) {
          reject(error)
          return
        }
        resolve(result.toString())
      })
    })
  }

  /**
   * Query for symbol for a given ERC20 asset.
   *
   * @param address - ERC20 asset contract address.
   * @returns Promise resolving to the 'symbol'.
   */
  async getTokenSymbol(address?: string): Promise<string> {
    const contract = address
      ? new this.web3.eth.Contract(FT_ABI, address, { from: this.account.address })
      : this.contract
    return new Promise<string>((resolve, reject) => {
      contract.methods.symbol().call((error: Error, result: BN | string) => {
        /* istanbul ignore if */
        if (error) {
          reject(error)
          return
        }
        resolve(result.toString())
      })
    })
  }
  /**
   * Query if a contract implements an interface.
   *
   * @param address - Asset contract address.
   * @param userAddress - The public address for the currently active user's account.
   * @returns Promise resolving an object containing the standard, decimals, symbol and balance of the given contract/userAddress pair.
   */
  async getDetails({ address, userAddress }: { address?: string; userAddress: string }): Promise<{
    standard: string
    symbol: string | undefined
    decimals: string | undefined
    balance: BN | undefined
  }> {
    const [decimals, symbol] = await Promise.all([this.getTokenDecimals(address), this.getTokenSymbol(address)])
    let balance
    if (userAddress) {
      balance = await this.getBalanceOf({ address, selectedAddress: userAddress })
    }
    return {
      decimals,
      symbol,
      balance,
      standard: 'ERC20',
    }
  }

  async transfer({
    address,
    from,
    to,
    amount,
    account,
    gas,
  }: {
    address: string
    from: string
    to: string
    account?: string
    amount: number | string
    gas?: number
  }) {
    const contract = new this.web3.eth.Contract(FT_ABI, address, { from: account || this.account.address })
    const amountBN = Web3.utils.toBN(Web3.utils.toWei(amount + ''))
    return contract.methods.transfer(to, amountBN).send({
      gas: gas || 1000000,
    })
  }

  async mint({ address, to, amount, account }: { account?: string; address?: string; to: string; amount: string }) {
    const contract = address
      ? new this.web3.eth.Contract(FT_ABI, address, { from: account || this.account.address })
      : this.contract
    // amount = this.web3.utils.toWei(amount + '')
    let amountBN = Web3.utils.toBN(Web3.utils.toWei(amount + ''))
    let gas = await contract.methods.mint(to, amountBN).estimateGas({ gas: 1000000 })
    return contract.methods.mint(to, amountBN).send({ gas: (gas * 1.1) | 0 })
  }

  async lock({ address, amount, account }: { account?: string; address?: string; amount: string }) {
    address = address || this.contract.options.address
    // amount = this.web3.utils.toWei(amount + '')
    const contract = address
      ? new this.web3.eth.Contract(FT_ABI, address, { from: account || this.account.address })
      : this.contract
    let locker = account
      ? new this.web3.eth.Contract(FT_LOCKER_ABI, this.lockerAddress, { from: account })
      : this.locker
    const amountBN = Web3.utils.toBN(Web3.utils.toWei(amount + ''))
    await contract.methods.increaseAllowance(locker.options.address, amountBN).send({ gas: 1000000 })
    let gas = await locker.methods.lock(address, amountBN).estimateGas({ gas: 1000000 })
    return locker.methods.lock(address, amountBN).send({ gas: (gas * 1.1) | 0 })
  }

  async unLock({ address, to, amount, account }: { account?: string; address?: string; to: string; amount: string }) {
    address = address || this.contract.options.address
    let locker = account
      ? new this.web3.eth.Contract(FT_LOCKER_ABI, this.lockerAddress, { from: account })
      : this.locker
    const contract = address
      ? new this.web3.eth.Contract(FT_ABI, address, { from: account || this.account.address })
      : this.contract
    const amountBN = Web3.utils.toBN(Web3.utils.toWei(amount + ''))
    let gas = await locker.methods.unlock(address, to, amountBN).estimateGas({ gas: 1000000 })
    return locker.methods.unlock(address, to, amountBN).send({ gas: (gas * 1.1) | 0 })
  }

  async getLockedNum({ address, account }: { address?: string; account: string }) {
    address = address || this.contract.options.address
    return this.locker.methods.lockedNum(address, account).call()
  }

  async getPastEvents({ address, fromBlock }: { address?: string; fromBlock: number }) {
    const contract = address
      ? new this.web3.eth.Contract(FT_ABI, address, { from: this.account.address })
      : this.contract
    return contract.getPastEvents('Transfer', {
      fromBlock,
      toBlock: fromBlock + 50000,
    })
  }

  async getPastLockEvents({ address, fromBlock }: { address?: string; fromBlock: number }) {
    return this.locker.getPastEvents('Lock', {
      fromBlock,
      toBlock: fromBlock + 50000,
    })
  }

  async getPastUnLockEvents({ address, fromBlock }: { address?: string; fromBlock: number }) {
    return this.locker.getPastEvents('UnLock', {
      fromBlock,
      toBlock: fromBlock + 50000,
    })
  }
}
